#include "arith.h"
#include "bwt.h"
#include "crc.h"
#include "error.h"
#include "gst.h"
#include "model.h"
#include "tag.h"
#include "defs.h"

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum action_t { PRINT_HELP, COMPRESS, DECOMPRESS };

const char file_id[] = "CHL!";

static FILE *f_in, *f_out;
static enum action_t action;
static uint32_t max_block_size;

static void parse_args(int argc, char **argv);
static void print_help(const char *);
static void encode(void);
static void decode(void);

int main(int argc, char **argv)
{
    error_set_level(1024);

    f_in = f_out = NULL;
    action = PRINT_HELP;
    max_block_size = 1024;

    parse_args(argc, argv);

    error_print(3, "max_block_size = %uK", max_block_size);
    max_block_size *= 1024;

    switch(action) {
        case PRINT_HELP:
            print_help(argv[0]);
            break;

        case COMPRESS:
        case DECOMPRESS:
            f_in = fopen(argv[argc-2], "rb");
            if(f_in == NULL) {
                error_print(0, "Can't open %s", argv[argc-2]);
                break;
            }

            f_out = fopen(argv[argc-1], "wb");
            if(f_out == NULL) {
                error_print(0, "Can't open %s", argv[argc-1]);
                break;
            }

            if(action == COMPRESS) encode();
            if(action == DECOMPRESS) decode();

            fclose(f_in);
            fclose(f_out);

            break;
    }

    return 0;
}

/*
 * print_help() - afiseaza help
 */
static void print_help(const char *name)
{
    fprintf(stderr, "Usage:\n");
    fprintf(stderr, "    %s [options] file_in file_out\n", name);
    fprintf(stderr, "\n");
    fprintf(stderr, "[options] can be any of the following\n");
    fprintf(stderr, "    -h       print this page\n");
    fprintf(stderr, "    -c       compress file_in to file_out\n");
    fprintf(stderr, "    -d       decompress file_in to file_out\n");
    fprintf(stderr, "    -v=n     level of verbosity\n");
    fprintf(stderr, "               -1 - quiet\n");
    fprintf(stderr, "                0 - adds critical errors\n");
    fprintf(stderr, "                1 - adds warnings\n");
    fprintf(stderr, "                2 - adds info about curent action\n");
    fprintf(stderr, "              >=3 - adds lots of debug texts\n");
    fprintf(stderr, "                default 1024 (big enough to output everything)\n");
    fprintf(stderr, "    -b=n     set the maximum block size in k\n");
    fprintf(stderr, "                1 <= n <= 1048576, default 1024\n");
    fprintf(stderr, "\n");
}

/*
 * parse_args() - o functie elementara de setare a anumitor optiuni
 */
static void parse_args(int argc, char **argv)
{
    int i;

    #define ERROR_PARSE() { \
            action = PRINT_HELP; \
            return; \
        }

    if(argc < 3) ERROR_PARSE();

    for(i = 1; i < argc-2; ++ i) {
        if(strcmp(argv[i], "-c") == 0) {
            action = COMPRESS;
        } else if(strcmp(argv[i], "-d") == 0) {
            action = DECOMPRESS;
        } else if(strcmp(argv[i], "-h") == 0) {
            action = PRINT_HELP;
        } else if(strncmp(argv[i], "-v=", 3) == 0) {
            error_set_level(atoi(argv[i]+3));
        } else if(strncmp(argv[i], "-b=", 3) == 0) {
            max_block_size = atoi(argv[i]+3);
            if(0 >= max_block_size || max_block_size > (1 << 20)) {
                fprintf(stderr, "Invalid value for MaxBlockSize\n");
                ERROR_PARSE();
            }
        } else {
            fprintf(stderr, "Unkown option \"%s\"\n", argv[i]);
            ERROR_PARSE();
        }
    }

    #undef ERROR_PARSE

    return;
}

/* functii locale, encode */
static void do_encode_flush(const uint8_t *buffer, uint32_t size)
{
    fwrite(buffer, 1, size, f_out);
}

static int do_compress_block(void)
{
    tag_block_t tag_block;
    long where, temp;
    uint8_t *t1; uint32_t s1;
    uint8_t *t2; uint32_t s2;
    uint32_t i;

    /* leave space for block tag */
    where = ftell(f_out);
    fwrite(&tag_block, sizeof(tag_block_t), 1, f_out);

    /* read and transform for AC */
    t1 = (uint8_t *)malloc(max_block_size);
    assert(t1 != NULL);
    s1 = fread(t1, 1, max_block_size, f_in);

    if(s1 == 0) return 0;

    tag_block.size_in = s1;
    tag_block.crc_in = crc(t1, s1);

    error_print(2, "tag_block.size_in = %u", tag_block.size_in);
    error_print(3, "tag_block.crc_in = %08X", tag_block.crc_in);

    bwt_forward(t1, s1, &t2, &s2, &tag_block.bwt_index);
    free(t1); t1 = t2; s1 = s2;
    error_print(3, "bwt crc = %08X", crc(t1, s1));
    
    mtf_forward(t1, s1, &t2, &s2);
    free(t1); t1 = t2; s1 = s2;
    error_print(3, "mtf crc = %08X", crc(t1, s1));

    /*
    rle_forward(t1, s1, &t2, &s2);
    free(t1); t1 = t2; s1 = s2;
    error_print(3, "rle crc = %08X", crc(t1, s1));
    error_print(3, "rle size = %u", s1);
    */

    /* gata, incepem AC */
    tag_block.arith_in = s1;

    arith_encode_init(do_encode_flush);
    model_init();

    for(i = 0; i < s1; ++ i) {
        uint32_t low, high, scale;

        model_insert_symbol(t1[i]);
        while(model_get_freq(&low, &high, &scale) != 0) {
            arith_encode(low, high, scale);
        }
    }

    arith_encode_finish();
    free(t1);

    tag_block.size_out = arith_encode_wrote();
    error_print(2, "tag_block.size_out = %u", tag_block.size_out);
    error_print(2, "ratio %.2f", tag_block.size_out*100./tag_block.size_in);

    /* scriem tagul blockului */
    temp = ftell(f_out);
    fseek(f_out, where, SEEK_SET);
    fwrite(&tag_block, sizeof(tag_block_t), 1, f_out);
    fseek(f_out, temp, SEEK_SET);
    error_print(3, "Wrote block tag");

    return 1;
}

/*
 * encode() - comprima f_in si scrie rezultatul in f_out
 */
void encode(void)
{
    tag_file_t tag_file;
    long where, temp;

    error_print(2, "Compressing...");

    /* leave space for file tag */
    where = ftell(f_out);
    fwrite(&tag_file, sizeof(tag_file_t), 1, f_out);

    /* start compressing */
    tag_file.no_blocks = 0;
    do {
        error_print(2, "tag_file.no_blocks = %u", tag_file.no_blocks);

        if(do_compress_block() != 0)
            tag_file.no_blocks ++;
    } while(feof(f_in) == 0);

    memcpy(&tag_file.id, file_id, 4);

    /* write the file tag */
    temp = ftell(f_out);
    fseek(f_out, where, SEEK_SET);
    fwrite(&tag_file, sizeof(tag_file_t), 1, f_out);
    fseek(f_out, temp, SEEK_SET);
    error_print(3, "Wrote file tag");
}

/* functii locale, decode */

static void do_decode_flush(uint8_t *buff, uint32_t *size)
{
    *size = fread(buff, 1, *size, f_in);
}

static void do_decompress_block(void)
{
    tag_block_t block_tag;
    uint8_t *t1; uint32_t s1;
    uint8_t *t2; uint32_t s2;
    uint32_t i;

    fread(&block_tag, 1, sizeof(tag_block_t), f_in);

    s1 = block_tag.arith_in;
    t1 = (uint8_t *)malloc(s1);
    assert(t1 != NULL);

    arith_decode_init(do_decode_flush, block_tag.size_out);
    model_init();

    for(i = 0; i < s1; ++ i) {
        uint32_t code, scale;
        uint32_t low, high;
        int byte;

        do {
            model_get_scale(&scale);
            arith_decode_get_code(&code, scale);
            /* fprintf(stderr, "code = %5u scale = %5u ", code, scale); */
            byte = model_insert_code(code);
            model_get_freq(&low, &high, &scale);
            /* fprintf(stderr, "low = %5u high = %5u\n", low, high); */
            arith_decode_remove(low, high, scale);
        } while(byte == -1);

        t1[i] = (uint8_t)byte;
    }

    error_print(2, "read %u byte(s)", block_tag.size_out);

    /*
    error_print(3, "rle size = %u", s1);
    error_print(3, "rle crc = %08X", crc(t1, s1));
    rle_reverse(t1, s1, &t2, &s2);
    free(t1); t1 = t2; s1 = s2;
    */

    error_print(3, "mtf crc = %08X", crc(t1, s1));
    mtf_reverse(t1, s1, &t2, &s2);
    free(t1); t1 = t2; s1 = s2;

    error_print(3, "bwt crc = %08X", crc(t1, s1));
    bwt_reverse(t1, s1, &t2, &s2, block_tag.bwt_index);
    free(t1); t1 = t2; s1 = s2;

    error_print(1, "s1 %c= block_tag.size_in", s1 == block_tag.size_in ? '=' : '!');
    error_print(1, "crc %c= block_tag.crc_in", crc(t1, s1) == block_tag.crc_in ? '=' : '!');

    fwrite(t1, 1, s1, f_out);
    error_print(2, "wrote %u byte(s)", s1);
    
    free(t1);
}

/*
 * decode() - decomprime f_in si scrie rezultatul in f_out
 */
void decode(void)
{
    tag_file_t file_tag;

    error_print(2, "Decompressing...");

    /* reading the tag */
    fread(&file_tag, 1, sizeof(tag_file_t), f_in);
    if(memcmp(&file_tag.id, file_id, 4) != 0)
        error_print(1, "tag_file.id != %s", file_id);

    while(file_tag.no_blocks --) {
        error_print(3, "tag_file.no_blocks = %u", file_tag.no_blocks);
        do_decompress_block();
    }
}
