#include "error.h"
#include "model.h"
#include "defs.h"
#include <assert.h>
#include <string.h>
#include <math.h>

#define LIST_SIZE 16                            /* o lista cu frecventele */
#define BUFF_SIZE 65536

#define CHAIN_NO        4
#define CHAIN_ORD    1024
#define CHAIN_PRUNE   400
#define CHAIN_SWITCH  192
#define CHAIN0_ORD     16
#define CHAIN0_PRUNE  400

typedef struct freq_t freq_t;

struct freq_t {
	/*
     * uint32_t seen;
     */
	uint32_t total;
	uint32_t freq[256];
};

static uint32_t chain[CHAIN_ORD][CHAIN_NO+1];    /* Markov chain (order 3) for 0, 1, 2 */
static uint32_t chain_prev;
static uint32_t chain0[CHAIN0_ORD][CHAIN_NO+1];  /* we keep 2 chains and choose one of them */
static uint32_t chain0_prev;

static freq_t order0;                            /* Markoff chain (order 0) for other symbols */
static uint32_t decode_step;

static uint8_t buff[BUFF_SIZE];
static uint32_t buff_size, buff_head, buff_half;

static uint32_t list_freq[LIST_SIZE][3];         /* low, high, scale */
static uint32_t list_head, list_tail, list_count;

/* functii pentru lista de frecvente */

static void list_add_freq(uint32_t l, uint32_t h, uint32_t s)
{
	assert(list_count < LIST_SIZE);

	/*
	error_print(3, "l = %d h = %d s = %d", l, h, s);
	*/

	list_freq[list_head][0] = l;
	list_freq[list_head][1] = h;
	list_freq[list_head][2] = s;

	list_count ++;
	if(++ list_head == LIST_SIZE)
		list_head = 0;
}

static void list_get_freq(uint32_t *l, uint32_t *h, uint32_t *s)
{
	assert(list_count > 0);

	*l = list_freq[list_tail][0];
	*h = list_freq[list_tail][1];
	*s = list_freq[list_tail][2];

	list_count --;
	if(++ list_tail == LIST_SIZE)
		list_tail = 0;
}

static void freq_add_to_list(const freq_t *f, int byte)
{
	int i;
	uint32_t s;

	for(s = 0, i = CHAIN_NO-1; i < byte; ++ i)
		s += f->freq[i];

	assert(f->freq[i] != 0);
	list_add_freq(s, s+f->freq[i], f->total);
}

/* functii pentru compresie/decompresie */

/*
 * model_init()
 * initializeaza modelul
 */
void model_init(void)
{
	uint32_t i;

	/* initializam lista de frecvente */
	list_head = list_tail = list_count = 0;

	/* initializam lantul Markoff-3 pentru 3 elemente */
	chain_prev = 0;
	for(i = 0; i < CHAIN_ORD; ++ i) {
		int j;
        chain[i][CHAIN_NO] = 0;
		for(j = 0; j < CHAIN_NO; ++ j) {
			chain[i][j] = CHAIN_NO - j;
            chain[i][CHAIN_NO] += chain[i][j];
        }
	}

    chain0_prev = 0;
    for(i = 0; i < CHAIN0_ORD; ++ i) {
        int j;
        for(j = 0; j < CHAIN_NO; ++ j)
            chain0[i][j] = 1;
        chain0[i][CHAIN_NO] = CHAIN_NO;
    }

	/* initializam order0 */
	order0.total = 0;
    for(i = CHAIN_NO-1; i < 256; ++ i) {
	    order0.freq[i] = (uint32_t)(exp((-(double)i+CHAIN_NO-1.) / 9.5)*123)+1;
        order0.total += order0.freq[i];
    }

	/* pentru decompresie */
	decode_step = 0;

	/* buffer-ul */
	buff_size = buff_head = 0;
    buff_half = 256;
}

/*
 * model_update()
 * face update la model
 */
static void model_update(uint8_t byte)
{
	uint32_t which = (uint32_t)byte >= CHAIN_NO-1 ? CHAIN_NO-1 : (uint32_t)byte;

	/* facem update la chain */
    chain[chain_prev][which] += 2;
    chain[chain_prev][CHAIN_NO] += 2;

    /* zero trebuie pretuit mai mult */
    {
        static int bla = 1;

        if(which == 0) {
            if(bla == 8) {
                chain[chain_prev][0] ++;
                chain[chain_prev][CHAIN_NO] ++;
                bla = 0;
            }
            bla ++;
        }
    }

	if(chain[chain_prev][CHAIN_NO] >= CHAIN_PRUNE) {
		int j;

		chain[chain_prev][CHAIN_NO] = 0;

		for(j = 0; j < CHAIN_NO; ++ j) {
			chain[chain_prev][j] = chain[chain_prev][j]/2 + chain[chain_prev][j]%2;
			chain[chain_prev][CHAIN_NO] += chain[chain_prev][j];
		}
	}

    chain_prev = (chain_prev + which*CHAIN_ORD) / CHAIN_NO;

    /* facem update la chain0 */
    chain0[chain0_prev][which] += 2;
    chain0[chain0_prev][CHAIN_NO] += 2;

    if(chain0[chain0_prev][CHAIN_NO] >= CHAIN0_PRUNE) {
        int j;

        chain0[chain0_prev][CHAIN_NO] = 0;

        for(j = 0; j < CHAIN_NO; ++ j) {
            chain0[chain0_prev][j] = chain0[chain0_prev][j]/2 + chain0[chain0_prev][j]%2;
            chain0[chain0_prev][CHAIN_NO] += chain0[chain0_prev][j];
        }
    }

    chain0_prev = (chain0_prev + which*CHAIN0_ORD) / CHAIN_NO;

	/* order0 */
	if(byte >= CHAIN_NO-1) {
		/* we may use freq_visit() but it's faster this way */
		order0.freq[byte] += 7;
		order0.total += 7;

		if(buff_size < BUFF_SIZE) {
			buff[buff_size ++] = byte;
		} else {
			order0.freq[buff[buff_head]] -= 6;
			order0.total -= 6;

            order0.freq[buff[buff_half]] -= 1;
            order0.total -= 1;

			buff[buff_head] = byte;
			if(++ buff_head == BUFF_SIZE)
				buff_head = 0;
            if(++ buff_half == BUFF_SIZE)
                buff_half = 0;
		}
	}

}

/*
 * model_get_freq()
 * intoarce un numar != 0 daca mai exista frecvente in lista
 * intervalul il intoarce prin intermediul parametrilor
 */
int model_get_freq(uint32_t *l, uint32_t *h, uint32_t *s)
{
	if(list_count == 0) return 0;
	list_get_freq(l, h, s);
	return 1;
}

/* functii pentru compresie */

/*
 * model_insert_symbol()
 * primeste ca parametru un byte si insereaza in lista una sau mai multe frecvente
 * a.i. caracterul sa poata fi decodificat
 */
void model_insert_symbol(uint8_t byte)
{
	uint32_t which = (uint32_t)byte >= CHAIN_NO-1 ? CHAIN_NO-1 : (uint32_t)byte;
	uint32_t s, i;

	/*
	error_print(3, "byte = %2x esc = %d", (int)byte, order0.freq[0]);
	error_print(3, "seen = %d total = %d", order0.seen, order0.total);
	*/

	/* facem update la chain3 */
    if(chain[chain_prev][CHAIN_NO] > CHAIN_SWITCH) {
        for(s = 0, i = 0; i < which; ++ i)
            s += chain[chain_prev][i];

        list_add_freq(s, s+chain[chain_prev][which],
                      chain[chain_prev][CHAIN_NO]);
    } else {
        for(s = 0, i = 0; i < which; ++ i)
            s += chain0[chain0_prev][i];

        list_add_freq(s, s+chain0[chain0_prev][which],
                      chain0[chain0_prev][CHAIN_NO]);
    }

	/* order 0 */
    if(byte >= CHAIN_NO-1)
        freq_add_to_list(&order0, byte);

/*
    if(byte >= CHAIN_NO-1) {
        static int t = 0;

        if(++ t % 300 == 0) {
            int j = 0;
            for(j = 0; j < 256; ++ j)
                printf("%6d", order0.freq[j]);
            printf(" %d\n", order0.total);
        }

    }
*/

	model_update(byte);
}

/* functii pentru decompresie */
/*
 * model_get_scale()
 * intoarce scale pentru simbolul curent
 */
void model_get_scale(uint32_t *scale)
{
	if(decode_step == 0) {
        if(chain[chain_prev][CHAIN_NO] > CHAIN_SWITCH)
		    *scale = chain[chain_prev][CHAIN_NO];
        else
            *scale = chain0[chain0_prev][CHAIN_NO];
    }
	else
		*scale = order0.total;
}

/*
 * model_insert_code()
 * determina carui caracter ii corespunde codul curent
 * returneaza -1 daca mai trebuie introduse coduri
 * altfel returneaza caracterul (ca int)
 */

int model_insert_code(uint32_t code)
{
	uint32_t low;
	int i;

	if(decode_step == 1) {
		assert(code < order0.total);

		low = 0; i = CHAIN_NO-1;
        	while(low + order0.freq[i] <= code)
			low += order0.freq[i ++];
		list_add_freq(low, low + order0.freq[i], order0.total);

		decode_step = 0;
		model_update((uint8_t)i);
		return i;
	}

    if(chain[chain_prev][CHAIN_NO] > CHAIN_SWITCH) {
        assert(code < chain[chain_prev][CHAIN_NO]);

	    low = 0; i = 0;
	    while(low + chain[chain_prev][i] <= code)
		   low += chain[chain_prev][i ++];
	    list_add_freq(low, low + chain[chain_prev][i],
	                  chain[chain_prev][CHAIN_NO]);
    } else {
        assert(code < chain0[chain0_prev][CHAIN_NO]);

        low = 0; i = 0;
        while(low + chain0[chain0_prev][i] <= code)
           low += chain0[chain0_prev][i ++];
        list_add_freq(low, low + chain0[chain0_prev][i],
                      chain0[chain0_prev][CHAIN_NO]);
    }

	if(i == CHAIN_NO-1) {
		decode_step = 1;
		return -1;
	}

    model_update((uint8_t)i);
	return i;
}
